/*-
 * Copyright 2007 Jason Harmening
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */

#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <limits.h>
#include <pthread.h>
#include <libgen.h>
#include <sys/errno.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <signal.h>

#include <string>
#include <iostream>
#include <sstream>
#include <fstream>
#include <list>
#include <iomanip>
#include <ext/stdio_filebuf.h>

#include "cx88_debug.h"
#include "cx88_capture_factory.h"
#include "cx88_mpeg_driver_factory.h"
#include "cx88_analog_driver_factory.h"

using namespace std;
using namespace __gnu_cxx;

#define DEFAULT_XML_CONFIG ".cx88.xml"
#define DEFAULT_RECORDING_FILE "/tmp/cx88.rec"
#define MAX_TCP_CONNECTIONS 8

extern char *optarg;

typedef struct
{
   cx88_driver *driver;
   double threshold;
   uint32_t lock_timeout_ms;
} cx88_scan_arg;

typedef struct
{
   cx88_driver *driver;
   capvec captures;
   char *device;
   tuner_config tunerconf;
} cx88_driver_entry;

typedef std::list<cx88_driver_entry> driverlist;

static int capture(
   cx88_driver_entry &entry,
   string &channel,
   string &recording_file,
   unsigned short port,
   bool auto_start,
   bool interactive,
   cx88_capture_params &params);

static void print_channel(void *arg, int error);
   
static void capture_complete(void *arg, int error);

static int scan(cx88_scan_arg &arg, cx88_capture_params &orig_params);

static int push_driver_entry(driverlist &drivers)
{
   try
   {
      cx88_driver_entry entry;
      entry.driver = NULL;
      entry.device = NULL;
      drivers.push_back(entry);
   }
   catch (...)
   {
      CX88ERR << "Unable to allocate memory for driver entry structure" << endl;
      return ENOMEM;  
   }
   return 0;
}

int main(int argc, char **argv)
{
   driverlist drivers;
   int error = 0, optch = -1;
   string xml_file = DEFAULT_XML_CONFIG;
   string channel = CX88_XML_CHANNEL_DEFAULT;
   string recording_file = DEFAULT_RECORDING_FILE;
   unsigned short remote_port = 0;
   bool auto_start = true;
   bool interactive = true;
   bool switch_audio = false;
   cx88_capture_params params;
   params.completion = NULL;
   params.completion_arg = NULL;
   params.channel_event = NULL;
   params.channel_arg = NULL;
   params.duration_s = 0;
   params.lock_timeout_ms = 1000;
   static const char *help = 
      "Options:\n"
      "\t-d <path>: [Required] Specify device from which to capture\n"
      "\t-c <channel | profile:channel>: [Optional] Specify initial channel, e.g. '7' or 'USA:7'\n"
      "\t-u <protocol://hostname:port>: [Optional] URL, Send capture over TCP or UDP to provided host\n"
      "\t-x <filename>: [Optional] Specify XML configuration file; default is "DEFAULT_XML_CONFIG"\n"
      "\t-t <filename>: [Optional] Specify libtuner configuration file for current device\n"
      "\t-o <string>: [Optional] Specify custom options in the form of <key0>=<value0>;<key1>=<value1>;...\n"
      "\t-n <value>: [Optional] Terminate capture after specified number of minutes\n"
      "\t-p <value>: [Optional] Specify port on which to listen for remote control connections\n"
      "\t-l <value> [Optional] Maximum channel lock timeout, in ms\n"
      "\t-q [Optional] Non-interactive capture--do not listen for input on stdin\n"
      "\t-h [Optional] Do not automatically start capturing\n"
      "\t-a [Optional] Force normal-mode (non-OSS) audio capture even when no capture URLs are specified\n"
      "\t-r [Optional] Specify target filename for recording; default is "DEFAULT_RECORDING_FILE"\n"
      "\t-? : Display this message\n";
   // Globally ignore SIGPIPE to keep from dying if the other end of a capture/connection drops out
   signal(SIGPIPE, SIG_IGN);
   error = push_driver_entry(drivers);
   while (!error && ((optch = getopt(argc, argv, "d:ac:u:x:t:n:o:p:l:qhr:?")) != -1))
   {
      switch (optch) 
      {
         case 'd':
         {
            char *device = strdup(optarg);
            if (device == NULL)
            {
               CX88ERR << "Unable to store device filename." << endl;
               error = ENOMEM;
            }
            if (drivers.back().device != NULL)
            {
               error = push_driver_entry(drivers);
            }
            if (!error)
            {
               drivers.back().device = device; 
               drivers.back().tunerconf.set_string(LIBTUNER_DOMAIN_KEY, basename(device));
            }
            break;
         }
         case 'c':
            try
            {
               channel = optarg;
            }
            catch (...)
            {
               CX88ERR << "Unable to store channel name.";
               error = ENOMEM;
            }
            break;
         case 'u':
         {
            cx88_capture *cap = cx88_capture_factory::create(optarg, drivers.back().tunerconf, error);
            if (cap == NULL)
            {
               CX88ERR << "Unable to store URL capture object." << endl;
               error = ENOMEM;
            }
            try
            {
               drivers.back().captures.push_back(cap);    
            }
            catch (...)
            {
               delete cap;
               CX88ERR << "Unable to store URL capture object." << endl;
               error = ENOMEM;
            }
            break;
         }
         case 'x':
            try
            {
               xml_file = optarg;
            }
            catch (...)
            {
               CX88ERR << "Unable to store XML configuration file name.";
               error = ENOMEM;
            }
            break;
         case 't':
            error = drivers.back().tunerconf.load_file(optarg);
            break;
         case 'n':
            params.duration_s = (unsigned long) (strtod(optarg, NULL) * 60);
            break;
         case 'o':
            drivers.back().tunerconf.load_string(optarg, ';');
            break;
         case 'p':
            remote_port = (unsigned short) (strtoul(optarg, NULL, 10));
            break;
         case 'l':
            params.lock_timeout_ms = (uint32_t) (strtoul(optarg, NULL, 10));
            break;
         case 'q':
            interactive = false;
            break;
         case 'h':
            auto_start = false;
            break;
         case 'a':
            switch_audio = true;
            break;
         case 'r':
            recording_file = optarg;
            break;
         case '?':
            cout << help;
            break;
         default:
            break;
      }
   }
   if (!error)
   {
      if (drivers.front().device == NULL)
      {
         CX88ERR << "Error: You must specify a device (-d)." << endl;
         error = EINVAL;
      }
   }
   cx88_xml_config xmlconf(xml_file.c_str(), error);
   if (!error)
   {
      cx88_driver *next = NULL;
      for (driverlist::reverse_iterator iter = drivers.rbegin(); iter != drivers.rend(); ++iter)
      {
         if (strstr(iter->device, "cx88mpeg") != NULL)
         {
            iter->driver = cx88_mpeg_driver_factory::create(
               xmlconf,
               iter->tunerconf,
               iter->device,
               iter->captures,
               error);
            if (!error && (iter->driver == NULL))
            {
               CX88ERR << "Unable to create cx88 MPEG driver" << endl;
               error = ENOMEM;  
            }
         }
         else if (strstr(iter->device, "cx88video") != NULL)
         {
            cx88_analog_driver *pip_driver = NULL;
            const char *pip_device = iter->tunerconf.get_string("pip");
            if (pip_device != NULL)
            {
               error = push_driver_entry(drivers);
               if (!error)
               {
                  cx88_driver_entry &entry = drivers.back();
                  entry.device = strdup(pip_device);
                  if (entry.device != NULL)
                  {
                     pip_driver = cx88_analog_driver_factory::create(
                        xmlconf,
                        entry.tunerconf,
                        entry.device,
                        entry.captures,
                        cx88_analog_driver::CX88_ANALOG_MODE_VIDEO,
                        NULL,
                        NULL,
                        error);
                  }
                  entry.driver = pip_driver;
               }
            }
            iter->driver = cx88_analog_driver_factory::create(
               xmlconf,
               iter->tunerconf,
               iter->device,
               iter->captures,
               cx88_analog_driver::CX88_ANALOG_MODE_VIDEO,
               dynamic_cast <cx88_analog_driver*>(next),
               pip_driver,
               error);
            if (!error && (iter->driver == NULL))
            {
               CX88ERR << "Unable to create cx88 video driver" << endl;
               error = ENOMEM;  
            }
         }
         else if (strstr(iter->device, "cx88audio") != NULL)
         {
            iter->driver = cx88_analog_driver_factory::create(
               xmlconf,
               iter->tunerconf,
               iter->device,
               iter->captures,
               (((switch_audio && (iter->captures.size() > 0)) || (!switch_audio && (iter->captures.size() == 0))) ? 
                  cx88_analog_driver::CX88_ANALOG_MODE_AUDIO_OSS :
                  cx88_analog_driver::CX88_ANALOG_MODE_AUDIO),
               dynamic_cast <cx88_analog_driver*>(next),
               NULL,
               error);
            if (!error && (iter->driver == NULL))
            {
               CX88ERR << "Unable to create cx88 audio driver" << endl;
               error = ENOMEM;  
            }
         }
         else
         {
            CX88ERR << "Invalid device filename" << endl;
            error = EINVAL;
         }
         for (capvec::size_type i = 0; i < iter->captures.size(); ++i)
         {
            iter->captures[i]->set_driver(iter->driver);
         }
         next = iter->driver;
      }
   }
   if (!error)
   {
      error = capture(drivers.front(), channel, recording_file, remote_port, auto_start, interactive, params);
   }
   for (driverlist::iterator iter = drivers.begin(); iter != drivers.end(); ++iter)
   {
      for (capvec::size_type i = 0; i < iter->captures.size(); ++i)
      {
         delete iter->captures[i];
      }
      delete iter->driver;
      free(iter->device);
   }
   return error;
}

static void capture_complete(void *arg, int error)
{
   int fd = *(reinterpret_cast <int*> (arg));
   static const char c[] = "q";
   write(fd, c, sizeof(c));
}

static void print_channel(void *arg, int error)
{
   cx88_scan_arg * scan_arg = reinterpret_cast<cx88_scan_arg *>(arg);
   sleep(1);
   double strength = (error ? 0.0 : scan_arg->driver->get_signal_strength());
   if (strength >= scan_arg->threshold)
   {
      ostringstream stream;
      stream << scan_arg->driver->get_channel().get_channel() << " (" << strength << "%) ";
      cout << stream.str().c_str();
   }
   else
   {
      cout << ". ";  
   }
   if (scan_arg->driver->get_channel().last_channel())
   {
      scan_arg->driver->stop();
   }
   else
   {
      string channel = CX88_XML_CHANNEL_UP;
      if (scan_arg->driver->set_channel(channel, scan_arg->lock_timeout_ms) != 0)
      {
         scan_arg->driver->stop();
      }
   }
   cout.flush();
}

static int capture(
   cx88_driver_entry &entry,
   string &channel,
   string &recording_file,
   unsigned short remote_port,
   bool auto_start,
   bool interactive,
   cx88_capture_params &params)
{
   cx88_driver *driver = entry.driver;
   cx88_file_capture *recording = NULL;
   int error = 0;
   fd_set fds;
   int server_sock = -1;
   bool halted = !auto_start;
   int maxfd = 0;
   int readyfds = 0;
   int controlfds[2];
   int client_socks[MAX_TCP_CONNECTIONS];
   stdio_filebuf<char> client_bufs[MAX_TCP_CONNECTIONS];
   streambuf *errbuf = cerr.rdbuf();
   streambuf *outbuf = cout.rdbuf();
   streambuf *logbuf = clog.rdbuf();
   controlfds[0] = -1;
   controlfds[1] = -1;
   for (size_t i = 0; i < MAX_TCP_CONNECTIONS; ++i)
   {
      client_socks[i] = -1;
   }
   bool keyboard_input = interactive;
   if (interactive)
   {
      cout << "Welcome to the cx88 capture app.  Enter '?' for a list of commands." << endl;
      maxfd = fileno(stdin) + 1;  
   }
   if (params.duration_s != 0)
   {
      if (pipe(controlfds))
      {
         CX88ERR << "Unable to create control file descriptor for scanning: " 
            << strerror(errno) << endl;  
         error = errno;
         goto cleanup;
      }
      if (controlfds[0] >= maxfd)
      {
         maxfd = controlfds[0] + 1;
      }
      params.completion = capture_complete;
      params.completion_arg = &(controlfds[1]);
   }
   if (remote_port != 0)
   {
      struct sockaddr_in server_addr;
      memset(&server_addr, 0, sizeof(server_addr));
      server_addr.sin_family = AF_INET;
      server_addr.sin_port = htons(remote_port);
      server_addr.sin_addr.s_addr = htonl(INADDR_ANY);
      if ((server_sock = socket(PF_INET, SOCK_STREAM, IPPROTO_TCP)) == -1)
      {
         CX88ERR << "Unable to create remote control socket: " << strerror(errno) << endl;
         error = errno;
         goto cleanup;
      }
      if (bind(server_sock, (struct sockaddr*) (&server_addr), sizeof(server_addr)) < 0) 
      {
         close(server_sock);
         CX88ERR << "Unable to bind remote control socket: " << strerror(errno) << endl;
         error = errno;
         goto cleanup;
      }
      if (listen(server_sock, 1) < 0)
      {
         close(server_sock);
         CX88ERR << "Unable to establish socket listener: " << strerror(errno) << endl;
         error = errno;
         goto cleanup;
      }
      if (server_sock >= maxfd)
      {
         maxfd = server_sock + 1;
      }
   }
   if (maxfd == 0)
   {
      CX88ERR << "You must either specify a time limit (-n) or remote control port (-p), or omit the -q option" << endl;
      error = EINVAL;
      goto cleanup;
   }
   if ((error = driver->set_channel(channel, params.lock_timeout_ms)))
   {
      CX88ERR << "Unable to set initial channel" << endl;
      goto cleanup;
   }
   if (!halted)
   {
      if ((error = driver->start(params)))
      {
         CX88ERR << "Unable to start driver" << endl;
         goto cleanup;
      }
   }
   do
   {
      if (keyboard_input)
      {
         cout << "> ";
         cout.flush();
      }
      if (readyfds <= 0)
      {
         FD_ZERO(&fds);
         if (interactive)
         {
            FD_SET(fileno(stdin), &fds);
         }
         if (server_sock != -1)
         {
            FD_SET(server_sock, &fds);
            for (size_t i = 0; i < MAX_TCP_CONNECTIONS; ++i)
            {
               if (client_socks[i] != -1)
               {
                  FD_SET(client_socks[i], &fds);
               }
            }
         }
         if (params.duration_s != 0)
         {
            FD_SET(controlfds[0], &fds);
         }
         if ((readyfds = select(maxfd, &fds, NULL, NULL, NULL)) <= 0)
         {
            continue;
         }
      }
      --readyfds;
      bool remote_input = false;
      keyboard_input = false;
      string line;
      vector<string> argv;
      try
      {
         if ((params.duration_s != 0) && FD_ISSET(controlfds[0], &fds))
         {
            FD_CLR(controlfds[0], &fds);
            char c = '\0';
            do
            {
               read(controlfds[0], &c, sizeof(c));
               if (c != '\0')
               {
                  line += c;
               }
            } while (c != '\0');
         }
         else if (FD_ISSET(fileno(stdin), &fds))
         {
            FD_CLR(fileno(stdin), &fds);
            getline(cin, line);
            if (cin.rdstate() != cin.goodbit)
            {
               interactive = false;
               continue;
            }
            keyboard_input = true;
         }
         else if ((server_sock != -1) && (FD_ISSET(server_sock, &fds)))
         {
            FD_CLR(server_sock, &fds);
            struct sockaddr_in client_addr;
            socklen_t client_addr_len;
            int client_sock = accept(server_sock, (struct sockaddr *)(&client_addr), &client_addr_len);
            if (client_sock == -1)
            {
               continue;
            }
            size_t i;
            for (i = 0; i < MAX_TCP_CONNECTIONS; ++i)
            {
               if (client_socks[i] == -1)
               {
                  client_socks[i] = client_sock;
                  new (&(client_bufs[i])) stdio_filebuf<char> (client_sock, ios_base::out);
                  break;
               }
            }
            if (i == MAX_TCP_CONNECTIONS)
            {
               close(client_sock);
               continue;
            }
            if (client_sock >= maxfd)
            {
               maxfd = client_sock + 1;
            }
            continue;
         }
         else
         {
            for (size_t i = 0; i < MAX_TCP_CONNECTIONS; ++i)
            {
               if ((client_socks[i] != -1) && (FD_ISSET(client_socks[i], &fds)))
               {
                  FD_CLR(client_socks[i], &fds);
                  char c;
                  ssize_t count = 0;
                  do
                  {
                     count = recv(client_socks[i], &c, sizeof(c), 0);
                     if ((c != '\0') && (c != '\n'))
                     {
                        line += c;
                     }
                  } while ((count > 0) && (c != '\0') && (c != '\n'));
                  if (count <= 0)
                  {
                     client_bufs[i].~stdio_filebuf();
                     client_socks[i] = -1;
                     continue;
                  }
                  cout.rdbuf(&(client_bufs[i]));
                  clog.rdbuf(&(client_bufs[i]));
                  cerr.rdbuf(&(client_bufs[i]));
                  remote_input = true;
                  break;
               }
            }
         }
         if (!remote_input)
         {
            cout.rdbuf(outbuf);
            clog.rdbuf(logbuf);
            cerr.rdbuf(errbuf);
         }
         size_t token_start = 0;
         size_t token_end;
         do
         {
            token_end = line.find_first_of(" \t", token_start, 2);
            if (token_end != token_start)
            {
               size_t count = token_end;
               if (token_end != string::npos)
               {
                  count = token_end - token_start;
               }
               argv.push_back(line.substr(token_start, count));
            }
            token_start = token_end + 1;
         } while (token_end != string::npos);
         if (!remote_input && (argv.size() == 1))
         {
            if (argv[0] == "c")
            {
               cout << "Enter channel:" << endl;
               getline(cin, line);
               argv.push_back(line);
            }
            else if (argv[0] == "s")
            {
               cout << "Enter signal strength threshold:" << endl;
               getline(cin, line);
               argv.push_back(line);
            }
            else if (argv[0] == "+")
            {
               cout << "Enter capture URL:" << endl;
               getline(cin, line);
               argv.push_back(line);   
            }
            else if (argv[0] == "-")
            {
               cout << "Enter capture number:" << endl;
               getline(cin, line);
               argv.push_back(line);   
            }
         }
      }
      catch (...)
      {
         CX88ERR << "Unable to allocate memory for command parsing" << endl;
         continue;
      }
      if (argv.size() == 0)
      {
         continue;
      }
      int cmd_error = 0;
      if (argv[0] == "c")
      {
         uint32_t lock_timeout_ms = params.lock_timeout_ms; 
         if (argv.size() < 2)
         {
            CX88ERR << "Channel name or number required\n" << endl;
            continue;
         }
         if (argv.size() > 2)
         {
            lock_timeout_ms = strtoul(argv[2].c_str(), NULL, 10);
         }
         if ((cmd_error = driver->set_channel(argv[1], lock_timeout_ms)) != 0)
         {
            CX88ERR << "Unable to change channel: " << strerror(cmd_error) << endl;
         }
      }
      else if (argv[0] == "i")
      {
         string str;
         if ((cmd_error = driver->get_info(str)) != 0)
         {
            CX88ERR << "Unable to get driver info: " << strerror(cmd_error) << endl;  
         }
         else
         {
            try
            {
               str += "Device:\n\t";
               str += entry.device;
               const char *driver_name = driver->get_name();
               if (driver_name != NULL)
               {
                  str += " (";
                  str += driver_name;
                  str += ")";
               }
               str += "\n";
               const cx88_source *source = driver->get_current_source();
               if ((source != NULL) && (source->name != NULL))
               {
                  str += "Input source:\n\t";
                  str += source->name;
                  str += "\n";
               }
            }
            catch (...)
            {
               CX88ERR << "Insufficient memory for device info string" << endl;
               continue;
            }
            capvec &captures = driver->get_captures();
            if (captures.size() > 0)
            {
               try
               {
                  ostringstream strstream;
                  strstream << "Captures:" << endl;
                  for (capvec::size_type i = 0; i < captures.size(); ++i)
                  {
                     strstream << "\t" << i << "\t" << captures[i]->get_id() << endl;
                  }
                  strstream << endl;
                  str += strstream.str();
               }
               catch (...)
               {
                  CX88ERR << "Insufficient memory for capture info string" << endl;
                  continue;
               }
            }
            cout << str.c_str() << endl;
         }
      }
      else if (argv[0] == "s")
      {
         if (argv.size() < 2)
         {
            CX88ERR << "Signal strength threshold % required\n" << endl;
            continue;
         }
         if (halted)
         {
            CX88ERR << "Unable to scan: driver is halted" << endl;
            continue;  
         }
         stringstream stream(argv[1]);
         cx88_scan_arg arg;
         arg.driver = driver;
         arg.lock_timeout_ms = params.lock_timeout_ms;
         if (argv.size() > 2)
         {
            arg.lock_timeout_ms = strtoul(argv[2].c_str(), NULL, 10);
         }
         stream >> arg.threshold;
         scan(arg, params);
      }
      else if (argv[0] == "h")
      {
         if (!halted)
         {
            driver->stop();  
            halted = true;
         }
         else
         {
            if ((cmd_error = driver->start(params)))
            {
               CX88ERR << "Unable to restart driver: " << strerror(cmd_error) << endl;
            }
            halted = false;
         }
      }
      else if (argv[0] == "+")
      {
         if (argv.size() < 2)
         {
            CX88ERR << "Capture URL required" << endl;
            continue;
         }
         cx88_capture *cap = cx88_capture_factory::create(argv[1].c_str(), entry.tunerconf, cmd_error);
         if ((cap == NULL) || (cmd_error != 0) || (driver->add_capture(*cap) != 0))
         {
            delete cap;
            CX88ERR << "Unable to create capture object." << endl;
            continue;
         }
         cap->set_driver(driver);
      }
      else if (argv[0] == "-")
      {
         if (argv.size() < 2)
         {
            CX88ERR << "Capture number required" << endl;
            continue;
         }
         cx88_capture *cap = driver->remove_capture(strtoul(argv[1].c_str(), NULL, 10));
         if (cap == NULL)
         {
            CX88ERR << "Capture " << argv[1] << " not found" << endl;
         }
         if (dynamic_cast<cx88_file_capture*>(cap) == recording)
         {
            recording = NULL;
         }
         delete cap;
      }
      else if (argv[0] == "r")
      {
         if (recording != NULL)
         {
            driver->remove_capture(recording);
            delete recording;
            recording = NULL;
            CX88LOG << "Stopped recording to " << recording_file.c_str() << endl;
         }
         else
         {
            string recording_id;
            try
            {
               recording_id = "[recording] ";
               recording_id += recording_file;
            }
            catch (...)
            {
               CX88ERR << "Insufficient memory for recording ID string" << endl;
            }
            recording = new cx88_file_capture(recording_id, recording_file.c_str(), cmd_error,
                                              entry.tunerconf.get_number<size_t>(CX88_BUFFER_KEY, CX88_DEFAULT_FILE_CACHE));
            if ((recording == NULL) || (cmd_error != 0) || (driver->add_capture(*recording) != 0))
            {
               delete recording;
               recording = NULL;
               CX88ERR << "Unable to add recording capture." << endl;
               continue;
            }
            CX88LOG << "Recording to " << recording_file.c_str() << endl;
         }
      }
      else if (argv[0] == "n")
      {
         if (argv.size() == 1)
         {
            const cx88_source *sources = driver->get_sources();
            if ((sources == NULL) || (sources->name == NULL))
            {
               cout << "No non-default input sources found" << endl;
               continue;
            }
            cout << "Available input sources:\n"
                 << "-----------------------------------------\n";
            size_t i = 0;
            do
            {
               cout << setw(30) << setfill(' ') << left << sources[i].name << setw(0) << sources[i].descr << endl;
            } while (sources[++i].name != NULL);
            cout << endl;
         }
         else if (driver->set_source(argv[1]))
         {
            CX88ERR << "Unable to change input source" << endl;
         }
      }
      else if (argv[0] == "?")
      {
         cout << "Interactive commands:\n"
              << "--------------------------------------------\n"
              << "c <channel> [<timeout>]       Change channel, optionally specifying lock timeout in ms\n"
              << "s <strength%> [<timeout>]     Scan for channels with at least the specified signal strength\n"
              << "i                             Get current tuning information\n"
              << "h                             Toggle start/stop of all captures\n"
              << "+ <URL>                       Add a capture using the specified URL\n"
              << "- <number>                    Remove capture by index (use 'i' to get indexed capture list)\n"
              << "r                             Toggle start/stop of recording to default recording file\n"
              << "n [<source>]                  Change the active input source, or list available sources\n";
         driver->print_help();
         cout << "\nCapture URLs:\n"
              << "--------------------------------------------\n";
         for (size_t i = 0; i < (sizeof(cx88_capture_factory::captures) / sizeof(cx88_capture_factory::captures[0])); ++i)
         {
            cout << setw(30) << setfill(' ') << left << cx88_capture_factory::captures[i].url_prefix
                 << setw(0) << cx88_capture_factory::captures[i].desc << endl;
         }
         cout << endl;
      }
      else if (argv[0] == "q")
      {
         break;
      }
      else
      {
         driver->command(argv);
      }
   } while (!error);
   driver->stop();

cleanup:
   for (size_t i = 0; i < MAX_TCP_CONNECTIONS; ++i)
   {
      if (client_socks[i] != -1)
      {
         client_bufs[i].~stdio_filebuf();
      }
   }
   if (server_sock != -1)
   {
      close(server_sock);
   }
   if (controlfds[0] != -1)
   {
      close(controlfds[0]);
   }
   if (controlfds[1] != -1)
   {
      close(controlfds[1]);
   }
   return error;
}

int scan(cx88_scan_arg &arg, cx88_capture_params &orig_params)
{
   int error = 0;
   string orig_channel;
   arg.driver->get_channel().get_channel_name(orig_channel);
   arg.driver->stop();
   int controlfds[2];
   if (pipe(controlfds))
   {
      CX88ERR << "Unable to create control file descriptor for scanning: " 
         << strerror(errno) << endl;  
      return errno;
   }
   cx88_capture_params scan_params;
   scan_params.completion = capture_complete;
   scan_params.completion_arg = &(controlfds[1]);
   scan_params.channel_event = print_channel;
   scan_params.channel_arg = &arg;
   scan_params.duration_s = 0;
   scan_params.lock_timeout_ms = orig_params.lock_timeout_ms;
   fd_set fds;
   FD_ZERO(&fds);
   FD_SET(controlfds[0], &fds);
   cout << "Scanning profile " << arg.driver->get_channel().get_profile_name() << ' ';
   string start_channel = CX88_XML_CHANNEL_FIRST;
   ofstream temperr("/dev/null");
   streambuf *origerr = cerr.rdbuf(temperr.rdbuf());
   error = arg.driver->set_channel(start_channel, arg.lock_timeout_ms);
   error = (error ? error : arg.driver->start(scan_params));
   if (error)
   {
      CX88ERR << "Unable to start driver for scan" << endl;
      close(controlfds[0]);
      close(controlfds[1]);
      return error;
   }
   char cmd = '\0';
   do
   {
      read(controlfds[0], &cmd, sizeof(cmd));
   } while (cmd != 'q');
   cout << "Finished scan." << endl;
   cerr.rdbuf(origerr);
   close(controlfds[0]);
   close(controlfds[1]);
   error = arg.driver->set_channel(orig_channel, orig_params.lock_timeout_ms);
   error = (error ? error : arg.driver->start(orig_params));
   if (error)
   {
      CX88ERR << "Unable to restart driver" << endl;
      return error;
   }
   return error;
}
